@@ -1,5 +1,5 @@
 package Data::TableReader;
-$Data::TableReader::VERSION = '0.004';
+$Data::TableReader::VERSION = '0.005';
 use Moo 2;
 use Try::Tiny;
 use Carp;
@@ -19,7 +19,6 @@ has fields              => ( is => 'rw', required => 1, coerce => \&_coerce_fiel
 sub field_list             { @{ shift->fields } }
 has field_by_name       => ( is => 'lazy' );
 has record_class        => ( is => 'rw', required => 1, default => sub { 'HASH' } );
-has filters             => ( is => 'rw' ); # list of coderefs to apply to the data
 has static_field_order  => ( is => 'rw' ); # force order of columns
 has header_row_at       => ( is => 'rw', default => sub { [1,10] } ); # row of header, or range to scan
 has header_row_combine  => ( is => 'rw', lazy => 1, builder => 1 );
@@ -383,7 +382,7 @@ sub _match_headers_dynamic {
 	my @unclaimed= grep { !$col_map{$_} } 0 .. $#$header;
 	if (@unclaimed) {
 		my $act= $self->on_unknown_columns;
-		my $unknown_list= join(', ', map $header->[$_], @unclaimed);
+		my $unknown_list= join(', ', map $self->_fmt_header_text($header->[$_]), @unclaimed);
 		$act= $act->($self, $header, \@unclaimed) if ref $act eq 'CODE';
 		if ($act eq 'use') {
 			$self->_log->('warn','%sIgnoring unknown columns: %s', $context, $unknown_list);
@@ -399,39 +398,28 @@ sub _match_headers_dynamic {
 	}
 	return [ map $col_map{$_}, 0 .. $#$header ];
 }
+# Make header string readable for log messages
+sub _fmt_header_text {
+	shift if ref $_[0];
+	$_[0] =~ s/[\0-\x1F]+/ /g;
+	$_[0] =~ s/^\s+//;
+	$_[0] =~ s/\s+$//;
+	$_[0];
+}
 
 
-has _iterator => ( is => 'rw', weak_ref => 1 );
 sub iterator {
-	my $self= shift;
-	if ($self->_iterator) {
-		# old one is still out there, so need a new data handle
-		my $data_iter= $self->decoder->iterator;
-		$data_iter->seek($self->_data_iter_record_start);
-		$self->_data_iter($data_iter);
-	}
-	$self->_iterator(my $i= $self->_build_iterator);
-	return $i;
-}
-
-sub _make_validation_callback {
-	my ($self, $field, $index)= @_;
-	my $t= $field->type;
-	ref $t eq 'CODE'? sub {
-		my $e= $t->($_[0][$index]);
-		defined $e? ([ $field, $index, $e ]) : ()
-	}
-	: $t->can('validate')? sub {
-		my $e= $t->validate($_[0][$index]);
-		defined $e? ([ $field, $index, $e ]) : ()
-	}
-	: croak "Invalid type constraint $t on field ".$field->name;
-}
-
-sub _build_iterator {
 	my $self= shift;
 	my $fields= $self->fields;
-	my $data_iter= $self->_table_found->{data_iter};
+	# Creating the record iterator consumes the data source's iterator.
+	# The first time after detecting the table, we continue with the same iterator.
+	# Every time after that we need to create a new data iterator and seek to the
+	# first record under the header.
+	my $data_iter= delete $self->_table_found->{data_iter};
+	unless ($data_iter) {
+		$data_iter= $self->decoder->iterator;
+		$data_iter->seek($self->_table_found->{first_record_pos});
+	}
 	my $col_map=   $self->_table_found->{col_map};
 	my $field_map= $self->_table_found->{field_map};
 	my @row_slice; # one column index per field, and possibly more for array_val_map
@@ -488,7 +476,6 @@ sub _build_iterator {
 		}
 	}
 	@arrayvals= reverse @arrayvals;
-	my @filters= @{ $self->filters || [] };
 	my ($n_blank, $first_blank, $eof);
 	my $sub= sub {
 		again:
@@ -525,15 +512,10 @@ sub _build_iterator {
 		# Collect all the array-valued fields from the tail of the row
 		$row->[$_->[0]]= [ splice @$row, $_->[1], $_->[2] ] for @arrayvals;
 		# stop here if the return class is 'ARRAY'
-		unless (@field_names) {
-			$_->($row) for @filters;
-			return $row;
-		}
+		return $row unless @field_names;
 		# Convert the row to a hashref
 		my %rec;
 		@rec{@field_names}= @$row;
-		# Apply any filters
-		$_->(\%rec) for @filters;
 		# Construct a class, if requested, else return hashref
 		return $class? $class->new(\%rec) : \%rec;
 	};
@@ -542,6 +524,20 @@ sub _build_iterator {
 	);
 }
 
+sub _make_validation_callback {
+	my ($self, $field, $index)= @_;
+	my $t= $field->type;
+	ref $t eq 'CODE'? sub {
+		my $e= $t->($_[0][$index]);
+		defined $e? ([ $field, $index, $e ]) : ()
+	}
+	: $t->can('validate')? sub {
+		my $e= $t->validate($_[0][$index]);
+		defined $e? ([ $field, $index, $e ]) : ()
+	}
+	: croak "Invalid type constraint $t on field ".$field->name;
+}
+
 sub _handle_blank_row {
 	my ($self, $first, $last)= @_;
 	my $act= $self->on_blank_row;
@@ -623,7 +619,7 @@ Data::TableReader - Extract records from "dirty" tabular data sources
 
 =head1 VERSION
 
-version 0.004
+version 0.005
 
 =head1 SYNOPSIS
 
@@ -653,15 +649,6 @@ but there's plenty of options to choose from...
       { name => 'zip', header => qw/zip\b|postal/i, type => US_Zipcode },
     ],
     
-    # could do this after extraction, but this fixes it before the type validation
-    filters => [
-       # they keep losing the leading zeroes on our zip codes. grr.
-       sub {
-         $_[0]{zip} =~ s/^(\d+)(-(\d+))?$/sprintf("%05d-%04d", $1, $3||0)/e;
-         return $_[0];
-       },
-    ],
-    
     # Our data provider is horrible; just ignore any nonsense we encounter
     on_blank_row => 'next',
     on_validation_fail => 'next',
@@ -723,13 +710,6 @@ order they were specified in the L</fields> specification.
 Setting it to anything else will return records created with
 C<< $record_class->new(\%fields); >>
 
-=head2 filters
-
-Array of filters which should be applied to the data records after they have
-been assembled but before they are passed to the record constructor (if any).
-Each element of this array should be a coderef which receives a hashref and
-returns it, possibly modified.
-
 =head2 static_field_order
 
 Boolean, whether the L</fields> must be found in columns in the exact order
